package org.eclipse.stardust.embedded;

import java.io.IOException;
import java.io.Reader;
import java.io.Serializable;
import java.sql.Clob;
import java.sql.Connection;
import java.sql.Date;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

import org.eclipse.stardust.common.config.CurrentVersion;
import org.eclipse.stardust.engine.api.query.ActivityFilter;
import org.eclipse.stardust.engine.api.query.ActivityInstanceQuery;
import org.eclipse.stardust.engine.api.query.ActivityInstances;
import org.eclipse.stardust.engine.api.query.ProcessInstanceQuery;
import org.eclipse.stardust.engine.api.query.ProcessInstances;
import org.eclipse.stardust.engine.api.runtime.ActivityInstance;
import org.eclipse.stardust.engine.api.runtime.ProcessInstance;
import org.eclipse.stardust.engine.api.runtime.ProcessInstanceState;

/**
 * 
 * User AbstractGernalIppContextTest if you have a real/remote Stardust engine running. 
 * Remember to adopt carnot.properties/Spring configurartion for remote connection!
 * 
 * @author 
 *
 */
public abstract class AbstractProcessContextTest extends AbstractEmbeddedIppContextTest {
	private static final Map<String, Object> EMPTY_DATA = Collections.unmodifiableMap(new HashMap<String, Object>(0));
	private static final long WAITFOR_DELAY = 333;
	private static final int WAITFOR_RETRIES = 33;

	protected Serializable getInDataValue(ActivityInstance pActivityInsatnce, String pDataId) {
		Serializable lActivityInstanceData = getWorkflowService().getInDataValue(pActivityInsatnce.getOID(), null,
				pDataId);
		logger.debug("In Data type '" + (lActivityInstanceData != null ? lActivityInstanceData.getClass() : "null") + "' | In Data value '"
				+ (lActivityInstanceData != null ? lActivityInstanceData : "null") + "'");
		return lActivityInstanceData;
	}
	
	@SuppressWarnings("unchecked")
	protected Map<String, Serializable> getStructuredInDataValue(ActivityInstance pActivityInsatnce, String pDataId) {
		Serializable lInDataValue = getInDataValue(pActivityInsatnce, pDataId);
		if (lInDataValue instanceof Map) {
			return (Map)lInDataValue;
		} else {
			fail("Expected "+pDataId+" to be a Map. But it was " + lInDataValue.getClass());
			return null;
		}
	}

	protected ProcessInstance startProcess(String pProcessId) {
		return getWorkflowService().startProcess(pProcessId, EMPTY_DATA, true);
	}

	protected ProcessInstance startProcess(String pProcessId, String pDataPath, Serializable pData) {
		Map<String, Serializable> lDatas = new HashMap<String, Serializable>(1);
		lDatas.put(pDataPath, pData);
		return getWorkflowService().startProcess(pProcessId, lDatas, true);
	}

	protected ProcessInstance startProcessAsync(String pProcessId) {
		return getWorkflowService().startProcess(pProcessId, EMPTY_DATA, false);
	}

	protected ActivityInstance waitForProcessToReachActivity(ProcessInstance pProcessInstance, String pActivityId) {
		long lProcessOid = pProcessInstance.getOID();
		ActivityInstanceQuery lQuery = ActivityInstanceQuery.findAll();
		lQuery.getFilter().add(ActivityInstanceQuery.PROCESS_INSTANCE_OID.isEqual(lProcessOid));
		lQuery.getFilter().add(ActivityFilter.forAnyProcess(pActivityId));

		for (int i = 0; i < WAITFOR_RETRIES; i++) {
			ActivityInstances lResult = getQueryService().getAllActivityInstances(lQuery);
			if (lResult.size() == 0) {
				logger.debug("ActivityInstance " + lProcessOid + " not found yet " + i + "/" + WAITFOR_RETRIES);
				try {
					Thread.sleep(WAITFOR_DELAY);
				} catch (InterruptedException e) {
					fail(e.getMessage());
				}
			} else {
				assertEquals(1, lResult.size());
				return lResult.get(0);
			}
		}

		lQuery = ActivityInstanceQuery.findAlive();
		lQuery.getFilter().add(ActivityInstanceQuery.PROCESS_INSTANCE_OID.isEqual(lProcessOid));
		ActivityInstances lResult = getQueryService().getAllActivityInstances(lQuery);
		StringBuilder sb = new StringBuilder("ProcessInstance ");
		sb.append(lProcessOid);
		sb.append(" didn't reach ");
		sb.append(pActivityId).append(" - Alive Activities for this PID:\n");
		for (ActivityInstance aActivityInstance : lResult) {
			sb.append(aActivityInstance.getActivity().getId()).append("\n");
		}
		
		fail(sb.toString());
		return null;
	}

	protected ProcessInstance waitForProcessToComplete(ProcessInstance pProcessInstance) {
		return waitForProcessToState(pProcessInstance, ProcessInstanceState.Completed);
	}

	protected ProcessInstance waitForProcessToState(ProcessInstance pProcessInstance, ProcessInstanceState... pState) {
		long lProcessOid = pProcessInstance.getOID();
		ProcessInstanceQuery lQuery = ProcessInstanceQuery.findInState(pState);
		lQuery.getFilter().add(ProcessInstanceQuery.OID.isEqual(lProcessOid));

		for (int i = 0; i < WAITFOR_RETRIES; i++) {
			ProcessInstances lResult = getQueryService().getAllProcessInstances(lQuery);
			if (lResult.size() == 0) {
				logger.debug("ProcessInstance " + lProcessOid + " didn't enter state " + pState + " yet " + i + "/"
						+ WAITFOR_RETRIES);
				try {
					Thread.sleep(WAITFOR_DELAY);
				} catch (InterruptedException e) {
					fail(e.getMessage());
				}
			} else {
				assertEquals(1, lResult.size());
				return lResult.get(0);
			}
		}

		fail("ProcessInstance " + lProcessOid + " didn't enter state " + pState[0].getName());
		return null;
	}

	protected void assertInDataValue(ActivityInstance pActivityInstance, String pDataId, Serializable pExpectedValue) {
		Serializable lActivityInstanceData = getInDataValue(pActivityInstance, pDataId);

		assertEquals(pExpectedValue, lActivityInstanceData);
	}

	@SuppressWarnings("unchecked")
	protected ActivityInstance completeActivityInstance(ActivityInstance pActivityInstance) {
		return completeActivityInstance(pActivityInstance, Collections.EMPTY_MAP);
	}

	protected ActivityInstance completeActivityInstance(ActivityInstance pActivityInstance, Map<String, ?> pData) {
		return getWorkflowService().activateAndComplete(pActivityInstance.getOID(), null, pData);
	}

	protected ActivityInstance completeActivityInstance(ActivityInstance pActivityInstance, String pDataId,
			Serializable pDataValue) {
		Map<String, Serializable> lDataMap = new HashMap<String, Serializable>(1);
		lDataMap.put(pDataId, pDataValue);

		return completeActivityInstance(pActivityInstance, lDataMap);
	}

	protected String clob2String(Clob pClob) throws SQLException, IOException {
		Reader r = pClob.getCharacterStream();
		int c;
		StringBuilder aSb = new StringBuilder();
		while ((c = r.read())!= -1) {
			aSb.append((char)c);
		}
		r.close();
		return aSb.toString();
	}
	
	protected void dumpPreparedStatementResult(PreparedStatement pPs, Object... pParams) throws SQLException, IOException {
		pPs.clearParameters();
		
		int i = 1;
		
		for (Object aParam : pParams) {
			pPs.setObject(i++, aParam);
		}
		
		printResultSet(pPs.executeQuery());
		
		pPs.clearParameters();
	}
	
	protected void printResultSet(ResultSet pResultSet) throws SQLException, IOException {
		int aColumnCount = pResultSet.getMetaData().getColumnCount();
		StringBuilder sb = new StringBuilder("Query result:\n||");
		int i;
		for (i=1; i <= aColumnCount; i++) {
			sb.append(pResultSet.getMetaData().getColumnName(i)).append(" || ");
		}
		while (pResultSet.next()) {
			sb.append("\n");
			for(i=1; i <= aColumnCount; i++) {
				Object aValue = pResultSet.getObject(i);
				if (aValue instanceof Clob) {
					aValue = clob2String((Clob)aValue);
				}
				sb.append(aValue).append(" | ");
			}
		}
		logger.info(sb.toString());
		pResultSet.first();
	}
	
	protected void dumpQueryResult(String pSql) throws SQLException, IOException {
		Connection lCon = getAuditTrailConnection();
		try {
			Statement lStmt = lCon.createStatement(ResultSet.TYPE_SCROLL_INSENSITIVE,
                    ResultSet.CONCUR_UPDATABLE);
			ResultSet lRs = lStmt.executeQuery(pSql);
			printResultSet(lRs);
		} finally {
			lCon.close();
		}
	}

	protected void dumpTable(String pTableName) throws SQLException, IOException {
		dumpQueryResult("SELECT * FROM " + pTableName);
	}
	

	@Override
	protected void onTearDown() throws Exception {
		if(true)return;
		Connection con = getAuditTrailConnection();
		try {
			Statement stmt = con.createStatement();
			ResultSet result = stmt.executeQuery("SELECT process_instance.oid, starttime,terminationtime,state, id " +
					" FROM process_instance " +
					" LEFT JOIN process_definition " +
					" ON process_instance.model = process_definition.model AND process_instance.processdefinition = process_definition.oid");
			
			while(result.next()) {
				System.out.println(result.getLong("oid") +"\t"+result.getString("id")+ "\t" + new Date(result.getLong("starttime")) + "\t" + new Date(result.getLong("terminationtime")) + "\t" + result.getInt("state"));
			}
			
		} finally {
			con.close();
		}
		
		printIPPVersion();
	}
	
	@Override
	protected void onSetUp() throws Exception {
		printIPPVersion();
	}

	private void printIPPVersion() {
		logger.info("Test runs with IPP " + CurrentVersion.getBuildVersionName());
	}
}
